import serial
import re
import serial.tools.list_ports
import random
import time


class ScoutSTX:
    def __init__(self, emul=False, keyword='ES', baudrate=9600, timeout=1, port=None) -> None:
        self.port = port
        self.keyword = keyword
        self.timeout = timeout
        self.baudrate = baudrate
        self.inst = None
        self.emul = emul
        self.inst_is_open = False
        if self.emul:
            self.emul_str = "[Emulation mode]:"
        else:
            self.emul_str = ""
    
    def connect(self):
        if not self.emul:
            if not self.inst_is_open:
                if self.port:
                    self.inst = serial.Serial(self.port, self.baudrate, timeout=self.timeout)
                else:
                    self.inst = self.port_search(self.keyword)

                if self.inst is not None:
                    print(f'Connected to ScoutSTX scale.')
                    self.inst_is_open = True
                    return True
                else:
                    print('ScoutSTX Connection failed.')
                    return False
            else:
                print('ScoutSTX is already connected.')
        else:
            print(f'{self.emul_str} Connected to ScoutSTX scale.')
            self.inst = 'emulated scale'
            self.inst_is_open = True
            return True
    
    def port_search(self, keyword):
        print('Searching for the device...')
        ports = serial.tools.list_ports.comports()
        for port, desc, hwid in sorted(ports):
            ser = serial.Serial(port, self.baudrate, timeout=self.timeout)
            ser.write(b'*IDN?\r\n')
            idn = ser.readline().strip().decode('ascii')
            
            if keyword in idn:
                print(f'"{keyword}" found in: {port}')
                return ser
            else:
                ser.close()
            
        print(f'"{keyword}" is not found on any port')
        return None
    
    def close_connection(self):
        if not self.emul:
            if self.inst is not None and self.inst.is_open:
                self.inst.close()
                print('ScoutSTX Connection closed.')
            else:
                print('No active ScoutSTX connection to close.')
        else:
            if self.inst is not None and self.inst_is_open:
                print(f'{self.emul_str} ScoutSTX Connection closed.')
        self.inst_is_open = False

    
    def is_connected(self):
        if self.inst is None:
            return False
        if self.emul:
            return self.inst_is_open
        else:
            return self.inst.is_open  
    
    def read_weight(self):
        # Send command to scale to read weight
        if self.is_connected():
            if not self.emul:
                time.sleep(1) # wait for the scale to settle
                self.inst.write(b'S\r\n')
                response = self.inst.readline().decode().strip()
                # Parse weight from response
                if response.startswith('S'):
                    match = re.search(r"[-+]?\d*\.\d+|\d+", response)
                    if match:
                        weight = float(match.group())
                    else:
                        print("Error parsing weight from scale's response")
                        return None
                else:
                    print("Error scale's response")
                    return None
            else:
                # Generate random weight
                min_value = 1.0
                max_value = 10.0
                weight = round(random.uniform(min_value, max_value), 2)
            
            return weight

        else:
            print(f'{self.emul_str} ScoutSTX Connection is not established.')
            return None

#==============================================================================    

# how to use this class
if __name__ == "__main__":
    
    scale = ScoutSTX(emul=True)
    #scale = ScoutSTX(port='COM7')
    
    # connect to the device
    scale.connect()

    wt = scale.read_weight_time()
    if wt: print(f"Weight: {wt[0]} g, at {wt[1]}")

    w = scale.read_weight()
    if w: print(f"Weight: {w} g")
    
    # close the connection
    scale.close_connection()


import nidaqmx
from YorkUphysLab.GwINSTEK import GPD as PSU
import time


class Actuator:
    def __init__(self, _DAQ_mame, _psu, emul=False, _max_time= 12) -> None:
        self.sdaq = _DAQ_mame
        self.psu = _psu
        self.max_time = _max_time
        self.max_pos = 100 # mm
        self.actuator_on = False
        self.emul = emul
        self.current_pos = 0
        if self.emul:
            self.emul_str = "[Emulation mode]:"
        else:
            self.emul_str = ""

    
    def set_position(self, pos):
        if not self.actuator_on:
            print(f'{self.emul_str} Actuator is switched OFF. Switch it ON first!')
            return None
        
        # when PSU connection is closed by another device:
        if not self.psu.is_connected():
            print("PSU Connection is not established. Switching off other devise sharing same PSU might have caused this. Switch on this decive and try again.")
            return None
        
        
        if 0 <= pos <= self.max_pos:
            Vctrl = -0.00004*pos*pos + 0.0528*pos + 0.1
        else:
            print(f'{self.emul_str} position ({pos} mm) out of range. use 0-{self.max_pos} mm')
            return None

        #if self.actuator_on:
        self.current_pos = self.get_position()
        required_time = abs(pos - self.current_pos)*self.max_time/self.max_pos + 1
            
        if not self.emul:
            with nidaqmx.Task() as task:
                task.ao_channels.add_ao_voltage_chan(f'{self.sdaq}/ao0','mychannel',0,5)
                task.write(Vctrl)
                task.stop()
        
        time.sleep(required_time)
        print(f'{self.emul_str} position set to {pos} mm')
        self.current_pos = pos    

    def get_position(self):
        if not self.actuator_on:
            print(f'{self.emul_str} Actuator is switched OFF! Switch it ON first!')
            return None
        
        # when PSU connection is closed by another device:
        if not self.psu.is_connected():
            print("PSU Connection is not established. Switching off other devise sharing same PSU might have caused this. Switch on this decive and try again.")
            return None
        
        #if self.actuator_on:
        if not self.emul:
            with nidaqmx.Task() as task:
                task.ai_channels.add_ai_voltage_chan(f'{self.sdaq}/ai0')
                Vadc = task.read()
            pos = 55.33223 - 31.31889*Vadc + 0.5730428*Vadc*Vadc
        else:
            pos = self.current_pos
        
        return round(pos,1)

    def switch_on(self):
        try:
            if not self.psu.is_connected():
                self.psu.connect()

            if self.psu.is_connected():
                self.psu.set_voltage(1, 12)
                self.psu.set_current(1, 0.5)
                self.psu.enable_output()
                self.actuator_on = True
                message = f'{self.emul_str} Actuator switched ON.'
            else:
                message = f'{self.emul_str} PSU Connection is not established.'

        except Exception as e:
            message = f'{self.emul_str} An error occurred: {str(e)}'
        
        print(message)
        
        return self.actuator_on
    
    def switch_off(self):
        self.psu.disable_output()
        #self.psu.close_connection()
        print(f'{self.emul_str} Actuator switched OFF.')
        self.actuator_on = False
        return True
#==============================================================================    

# how to use this class
if __name__ == "__main__":
    
    DAQ_mame = 'SDAQ-25'
    # create power-supply and actuator objects
    psu = PSU.GPD3303D(emul=False)
    actuator = Actuator(DAQ_mame, psu, emul=False)
    
    #psu.connect()
    actuator.switch_on()
    actuator.set_position(100)
    print(f'position moved to = {actuator.get_position()} mm')
        
    time.sleep(2)
    
    #actuator.switch_off()

    '''
        SenorDAQ terminals:
        1:  p0.0
        2:  p0.1
        3:  p0.2
        4:  p0.3
        5:  GND
        6:  +5V
        7:  PFIO
        8:  GND
        9:  AO0  ----> 
        10: GND
        11: AI0  ----> 
        12: AI1
    '''


import pyvisa
import numpy as np
import time
import logging

#----------------------------------------------
logging.getLogger().setLevel(logging.INFO)

class TBS1000:
    """
    Represents a Tektronix TBS1000 oscilloscope.

    Attributes:
        keyword (str): The keyword used to identify the oscilloscope.
        timeout (int): The timeout value for communication with the oscilloscope.
        encoding (str): The encoding used for communication with the oscilloscope.
        read_termination (str): The read termination character for communication with the oscilloscope.
        write_termination (str): The write termination character for communication with the oscilloscope.
        inst (visa.Resource): The instrument resource representing the connected oscilloscope.
        is_open (bool): Indicates whether the oscilloscope connection is open or closed.
        total_time (flote): total time span of the waveform in seconds
    """
    
    def __init__(self, keyword='TBS', timeout=10000, encoding = 'latin_1', read_termination = '\n') -> None:
        self.keyword = keyword
        self.timeout = timeout
        self.encoding = encoding
        self.read_termination = read_termination
        self.write_termination = None

        self.inst = None
        self.is_open = False

        self.total_time = 0
        self.record = 0
        

    #----------------------------------------------
    def connect(self):
        rm = pyvisa.ResourceManager()
        resources_list = rm.list_resources()
        
        for re in resources_list:
            if 'USB' in re:
                dev = rm.open_resource(re)
                if self.keyword in dev.query('*idn?'):
                    self.inst = dev
                    break
                else:
                    dev.close()
        if self.inst:
            logging.info(f"Tektronix TBS scope found: {self.inst.query('*idn?')}")
            self.is_open = True
            self.inst.timeout = self.timeout # ms
            self.inst.encoding = self.encoding
            self.inst.read_termination = self.read_termination
            self.inst.write_termination = self.write_termination
            self.inst.write('*cls') # clear Event Status Register (ESR)
            #self.config()
            return True
        else:
            logging.info('No Tektronix TBS scope was found!')
            return False
    
    #----------------------------------------------
    def config(self, hscale='5E-3', ch1scale='50E-3', ch2scale='2', trig='CH2'):
        """
        Configures the Tektronix TBS oscilloscope with the specified settings.

        Args:
            hscale (str): Horizontal scale (time/division) setting in seconds. Default is '5E-3'.
            ch1scale (str): CH1 vertical scale (voltage/division) setting in seconds. Default is '50E-3'.
            ch2scale (str): CH2 vertical scale (voltage/division) setting in seconds. Default is '2'.
            trig (str): Trigger source setting. Default is 'CH2'. Valid options are CH1 or CH2

        Returns:
            bool: True if the oscilloscope is successfully configured, False otherwise.

        Raises:
            TypeError: If any of the arguments are not of type str.
        """
        if not all(isinstance(arg, str) for arg in (hscale, ch1scale, ch2scale, trig)):
                raise TypeError("All arguments must be of type string.")

        # Rest of the code...
        if self.is_connected():
            self.inst.write('*rst')  # reset the instrument to a known state.
            r = self.inst.query('*opc?')  # queries the instrument to check if it has completed the previous operation.
            self.inst.write('autoset EXECUTE')  # autoset: automatically adjusts the oscilloscope's settings based on the input signal
            r = self.inst.query('*opc?')
            self.inst.write(f'HORIZONTAL:MAIN:SCALE {hscale}') # set horizontal scale (time/division)
            r = self.inst.query('*opc?')
            self.inst.write('CH1:COUPLING AC')
            r = self.inst.query('*opc?')
            self.inst.write(f'CH1:SCALE {ch1scale}') # set ch1 vertical scale (voltage/division)
            r = self.inst.query('*opc?')
            self.inst.write(f'CH2:SCALE {ch2scale}') # set ch2 vertical scale (voltage/division)
            r = self.inst.query('*opc?')
            self.inst.write(f'TRIGGER:MAIN:EDGE:SOURCE {trig}') # set trigger source to channel 2
            r = self.inst.query('*opc?')
            return True
        else:
            logging.info('Tektronix TBS scope is not connected!')
            return False


    #----------------------------------------------
    def close(self):
        if self.is_connected():
            self.inst.close()
            self.is_open = False
            logging.info('Tektronix TBS scope connection is closed.')
        else:
            logging.info('Tektronix TBS scope is not connected!')

    #----------------------------------------------
    def is_connected(self):
        return self.is_open

    #----------------------------------------------
    def get_idn(self):
        if self.is_connected():
            return self.inst.query('*idn?')
        else:
            return None
    
    #----------------------------------------------
    def get_period(self, channel):
            """
            Get the period of the waveform on the specified channel.
            
            Args:
                channel (int): The channel number (1 or 2).
            Returns:
                float: The measured period in seconds.
            Raises:
                ValueError: If the channel number is not 1 or 2.
            """
            
            if channel not in [1,2]:
                raise ValueError("Channel must be 1 or 2")
            
            self.inst.write('MEASUrement:IMMed:TYPE PERiod')
            self.inst.write(f'MEASUrement:IMMed:SOUrce CH{channel}')
            measured_period = self.inst.query('MEASUrement:IMMed:VALue?')
            
            return float(measured_period) # in seconds
    
    #----------------------------------------------
    def get_frequency(self, channel):
            """
            Get the frequency of the waveform on the specified channel.
            
            Args:
                channel (int): The channel number (1 or 2).
            Returns:
                float: The measured frequency in Hz.
            Raises:
                ValueError: If the channel number is not 1 or 2.
            """
            
            if channel not in [1,2]:
                raise ValueError("Channel must be 1 or 2")
            
            self.inst.write('MEASUrement:IMMed:TYPE FREQUENCY')
            self.inst.write(f'MEASUrement:IMMed:SOUrce CH{channel}')
            measured_freq = self.inst.query('MEASUrement:IMMed:VALue?')
            
            return float(measured_freq) # in seconds

    def get_amplitude(self, channel):
            """
            Get the amplitude of the waveform on the specified channel.
            
            Args:
                channel (int): The channel number (1 or 2).
            Returns:
                float: The measured amplitude in V.
            Raises:
                ValueError: If the channel number is not 1 or 2.
            """
            
            if channel not in [1,2]:
                raise ValueError("Channel must be 1 or 2")
            
            self.inst.write('MEASUrement:IMMed:TYPE AMplitude')
            self.inst.write(f'MEASUrement:IMMed:SOUrce CH{channel}')
            measured_amplitude = self.inst.query('MEASUrement:IMMed:VALue?')
            
            return float(measured_amplitude) # in seconds
    
    def get_phase(self, channel):
            """
            Get the phase difference from the selected waveform to the designated waveform
            
            Args:
                channel (int): The channel number (1 or 2).
            Returns:
                float: The measured phase in degrees.
            Raises:
                ValueError: If the channel number is not 1 or 2.
            """
            
            if channel not in [1,2]:
                raise ValueError("Channel must be 1 or 2")
            
            self.inst.write(f'MEASUrement:IMMed:SOUrce1 CH1')
            self.inst.write(f'MEASUrement:IMMed:SOUrce2 CH2')
            self.inst.write('MEASUrement:IMMed:TYPE PHAse')
            measured_phase = self.inst.query('MEASUrement:IMMed:VALue?')
            
            if channel == 1:
                return float(measured_phase) # in degrees
            else:
                return -1*float(measured_phase) # in degrees
    #----------------------------------------------
    def get_data(self, channel):
        """
        Retrieves the scaled waveform data from the oscilloscope for the specified channel.

        Args:
            channel (int): The channel number (1 or 2) for which to retrieve the data.

        Returns:
            tuple: A tuple containing the following elements:
                - scaled_time (numpy.ndarray): An array of scaled time values in milliseconds.
                - scaled_wave (numpy.ndarray): An array of scaled waveform values.
        """
        if not self.is_connected():
            return None
        if channel not in [1, 2]:
            raise ValueError("Channel must be 1 or 2")

        scaled_wave = []
        
        """
        self.inst.write('*rst')  # reset the instrument to a known state.
        r = self.inst.query('*opc?')  # queries the instrument to check if it has completed the previous operation.
        self.inst.write('autoset EXECUTE')  # autoset: automatically adjusts the oscilloscope's settings based on the input signal
        r = self.inst.query('*opc?')
        """
        
        # io config
        self.inst.write('header 0')
        self.inst.write('data:encdg RIBINARY')
        self.inst.write(f'data:source CH{channel}')
        self.inst.write('data:start 1')  # first sample
        self.record = int(self.inst.query('wfmpre:nr_pt?'))  # number of samples
        self.inst.write(f'data:stop {self.record}')  # last sample
        self.inst.write('wfmpre:byt_nr 1')  # 1 byte per sample
        # acq config
        self.inst.write('acquire:state 0')  # stop data acquisition
        self.inst.write('acquire:stopafter SEQUENCE')  # sets the acquisition mode to 'SEQUENCE': acquires a single waveform and then stops
        self.inst.write('acquire:state 1')  # run
        
        # data query
        bin_wave = self.inst.query_binary_values('curve?', datatype='b', container=np.array)
        tscale = float(self.inst.query('wfmpre:xincr?'))  # retrieve scaling factors
        tstart = float(self.inst.query('wfmpre:xzero?'))
        vscale = float(self.inst.query('wfmpre:ymult?'))  # volts / level
        voff = float(self.inst.query('wfmpre:yzero?'))  # reference voltage
        vpos = float(self.inst.query('wfmpre:yoff?'))  # reference position (level)
        
        # error checking
        r = int(self.inst.query('*esr?'))
        if r != 0b00000000:
            logging.info('event status register: 0b{:08b}'.format(r))
        r = self.inst.query('allev?').strip()
        if 'No events' not in r:
            logging.info(f'all event messages: {r}')

        total_time = tscale * self.record  # create scaled vectors
        tstop = tstart + total_time
        scaled_time = np.linspace(tstart, tstop, num=self.record, endpoint=False) * 1000  # time in ms

        unscaled_wave = np.array(bin_wave, dtype='double')  # data type conversion
        scaled_wave = (unscaled_wave - vpos) * vscale + voff

        self.total_time = total_time
        return scaled_time, scaled_wave
    
    #----------------------------------------------
    def get_data2(self):
        """
        Retrieves waveform data from the Tektronix TBS1000 oscilloscope.

        Returns:
            tuple: A tuple containing the following elements:
                - scaled_time (numpy.ndarray): An array of scaled time values in milliseconds.
                - scaled_wave[0] (numpy.ndarray): An array of scaled waveform data for channel 1.
                - scaled_wave[1] (numpy.ndarray): An array of scaled waveform data for channel 2.
        """
        if not self.is_connected():
            return None

        scaled_wave = []

        # acq config
        self.inst.write('acquire:state RUN')  # RUN acquisition
        self.inst.write('acquire:stopafter SEQUENCE')  # sets the acquisition mode to 'SEQUENCE': acquires a single waveform and then stops
        self.inst.write('acquire:state 0')  # stop data acquisition

        # io config
        self.inst.write('header 0')
        self.inst.write('data:encdg RIBINARY')

        for channel in [1,2]:
            # ch1
            self.inst.write(f'data:source CH{channel}')
            self.inst.write('data:start 1')  # first sample
            self.record = int(self.inst.query('wfmpre:nr_pt?'))  # number of samples
            self.inst.write(f'data:stop {self.record}')  # last sample
            self.inst.write('wfmpre:byt_nr 1')  # 1 byte per sample

            # data query
            bin_wave = self.inst.query_binary_values('curve?', datatype='b', container=np.array)
            tscale = float(self.inst.query('wfmpre:xincr?'))  # retrieve scaling factors
            tstart = float(self.inst.query('wfmpre:xzero?'))
            vscale = float(self.inst.query('wfmpre:ymult?'))  # volts / level
            voff = float(self.inst.query('wfmpre:yzero?'))  # reference voltage
            vpos = float(self.inst.query('wfmpre:yoff?'))  # reference position (level)

            # error checking
            r = int(self.inst.query('*esr?'))
            if r != 0b00000000:
                logging.info('event status register: 0b{:08b}'.format(r))
            r = self.inst.query('allev?').strip()
            if 'No events' not in r:
                logging.info(f'all event messages: {r}')

            total_time = tscale * self.record  # create scaled vectors
            tstop = tstart + total_time
            scaled_time = np.linspace(tstart, tstop, num=self.record, endpoint=False) * 1000  # time in ms

            unscaled_wave = np.array(bin_wave, dtype='double')  # data type conversion
            scaled_wave.append((unscaled_wave - vpos) * vscale + voff)

            self.total_time = total_time

        self.inst.write('acquire:state RUN')  # RUN acquisition

        return scaled_time, scaled_wave[0], scaled_wave[1]
    #----------------------------------------------
    def shift_phase(self, scaled_time, waveform_1, waveform_2, phi):
        """
        Apply phase shift to waveform_2 based on the given phi value. To keep the length of the waveforms the same, the shifted waveforms and scaled_time are truncated.
        
        Parameters:
        scaled_time (array-like): Array of scaled time values.
        waveform_1 (array-like): Array of waveform 1 values.
        waveform_2 (array-like): Array of waveform 2 values.
        phi (float): Phase shift value in degrees.
        
        Returns:
        tuple: A tuple containing the shifted scaled_time, waveform_1, and shifted waveform_2.
        """
        
        _, phase = divmod(phi, 360)    
        # get the period of waveform 2
        period = self.get_period(channel=2)
        samples_in_period = int(period/self.total_time * len(waveform_2)) # number of samples in one period
        shift_samples = int((phase / 360) * samples_in_period) # number of samples to shift

        if shift_samples == 0:
            return scaled_time, waveform_1, waveform_2
        else:
            return scaled_time[:-1*shift_samples], waveform_1[:-1*shift_samples], waveform_2[shift_samples:]
    
    #----------------------------------------------
    def shift_phase2(self, scaled_time, ref_channel, phi_shift):
        """
        Constructing an internal reference waveform with the given phase shift w.r.t the main waveform based on the given parameters.

        Parameters:
        scaled_time (float): The scaled time values for the waveform. Need to convert to sec.
        ref_channel (str): The reference channel for obtaining the frequency.
        phi_shift (float): The phase shift value in degrees.

        Returns:
        numpy.ndarray: The constructed waveform with the shifted phase.
        """
        
        freq = self.get_frequency(ref_channel)
        
        # constructing a pure internal ref. wave
        internal_ref_waveform = np.cos(2*np.pi*freq*scaled_time*0.001 + np.radians(phi_shift))

        return internal_ref_waveform

#==============================================================================
# how to use this class
if __name__ == '__main__':
    # create a scope object
    scope = TBS1000()
    # connect to the scope
    if scope.connect():
        scope.config(hscale='5E-3', ch1scale='50E-3', ch2scale='2', trig='CH2')
    
    color = {1:'orange', 2:'blue', 'mix':'red'}
    
    #"""
    try:
        # get data from the scope
        scaled_time, scaled_wave_1, scaled_wave_2 = scope.get_data2()
        
        phi = 0
        #stime, wf1, wf2_shifted = scope.shift_phase(scaled_time, scaled_wave_1, scaled_wave_2, phi)
        wf2_shifted = scope.shift_phase2(scaled_time, ref_channel=2, phi_shift = 330)

        mix_wf = [x*y for x,y in zip(scaled_wave_1,wf2_shifted)]

        avg_mix_wf = np.average(mix_wf)*1000 # in mV
        
        # --plotting
        #'''
        import pylab as pl
        #pl.plot(scaled_time, scaled_wave_1, label=f'Ch 1', color=color[1])
        #y_max = max(scaled_wave_1)

        pl.plot(scaled_time, scaled_wave_1, label=f'Ch 1', color=color[1])
        pl.plot(scaled_time, wf2_shifted, label=f'Ch 2', color=color[2])
        #pl.plot(stime, mix_wf, label=f'Mix', color=color['mix'])
        
        y_max = max(max(scaled_wave_1), max(wf2_shifted))
        #y_max = max(max(wf1), max(wf2_shifted), max(mix_wf))
       
        pl.ylim(top=y_max*1.5)
        pl.xlabel('time [ms]') # x label
        pl.ylabel('voltage [v]') # y label
        # Add legend
        pl.legend(loc='upper right')
        
        pl.rc('grid', linestyle=':', color='gray', linewidth=1)
        pl.grid(True)
        pl.title(f'Lock-in Output: {round(avg_mix_wf,2)} mV,  $\Delta\phi: {phi}\degree$', fontsize = 10)
        #'''
        
    except ValueError as e:
        print(e)
    

    scope.close()
    
    print("\nlook for plot window...")
    pl.show()
    print("\nend of demonstration")
    
    
    scope.close()



import YorkUphysLab.GwINSTEK.GPD as PSU
import time


class HV_control:
    def __init__(self, _psu, emul=False) -> None:
        self.psu = _psu
        self.HV_on = False
        self.emul = emul
        if self.emul:
            self.emul_str = "[Emulation mode]:"
        else:
            self.emul_str = ""

    def switch_on(self):
        try:
            if not self.psu.is_connected():
                self.psu.connect()

            if self.psu.is_connected():
                self.psu.set_voltage(1, 12)
                self.psu.set_current(1, 0.5)
                self.psu.set_voltage(2, 1)
                self.psu.set_current(2, 0.01)
                self.psu.enable_output()
                self.HV_on = True
                message = f'{self.emul_str} HV switched ON.'
            else:
                message = f'{self.emul_str} PSU Connection is not established.'
                
        except Exception as e:
            message = f'{self.emul_str} An error occurred: {str(e)}'
        
        print(message)
        return self.HV_on
    
    def switch_off(self):
        self.psu.disable_output()
        #self.psu.close_connection()
        print(f'{self.emul_str} HV switched OFF.')
        self.HV_on = False
        return True
    
    def set_hv(self, voltage):
        if not self.HV_on:
            print(f'{self.emul_str} HV is switched OFF! Switch it ON first!')
            return None
        
        # when PSU connection is closed by another device:
        if not self.psu.is_connected():
            print("PSU Connection is not established. Switching off other devise sharing same PSU might have caused this. Switch on this decive and try again.")
            return None
        
        if 0 <= voltage <= 3:
            # Vctrl equation is a linear fit of the data below
            Vctrl = 1.6489*voltage + 0.0595
            Vctrl = 5 if Vctrl >= 5 else round(Vctrl, 1)

            #if self.HV_on:
            self.psu.set_voltage(2, Vctrl)
            print(f'{self.emul_str} HV voltage set to {voltage} kV')
            return True
            
        else:
            print(f'{self.emul_str} Requested High voltage out of range. Use 0 -3 kV')
            return None
    
    def get_hv(self):
        
        if not self.HV_on:
            print(f'{self.emul_str} HV is switched OFF! Switch it ON first!')
            return None
        
        # when PSU connection is closed by another device:
        if not self.psu.is_connected():
            print("PSU Connection is not established. Switching off other devise sharing same PSU might have caused this. Switch on this decive and try again.")
            return None
        
        # the output must be enabled to read the voltage
        actual_voltage = self.psu.get_voltage(2)
        #print(f'{self.emul_str} actual_voltage = {actual_voltage} V')
        actual_HV = (actual_voltage - 0.0595)/1.6489
        return round(actual_HV, 2)
        
#==============================================================================    

# how to use this class
if __name__ == "__main__":
        
    try:
        #psu = PSU.GPD3303D(port='COM8')
        psu = PSU.GPD3303D(emul=True)
    except:
        print('No PSU found, or device is not connected')
        exit()
    
    psu.connect()

    if psu.is_connected():
        HV = HV_control(psu, emul=True)
        
        HV.switch_on()
        HV.set_hv(1.5)
        time.sleep(2)
        actual_HV = HV.get_hv()
        print(f'Actual HV = {actual_HV} kV')
        psu.disable_output()
        psu.close_connection()
    
    else:
        print('PSU not connected')


'''
# high voltage callibration
Vctrl	V kV(cur)
0.1,	0.025127956
0.2,	0.085740858
0.3,	0.146705996
0.4,	0.207464024
0.5,	0.26828344
0.6,	0.328773566
0.7,	0.389337894
0.8,	0.449891494
1.3,	0.753993044
1.8,	1.053181468
2.3,	1.354698166
2.8,	1.654354748
3.3,	1.955916146
3.8,	2.276596628
4.3,	2.578250108
4.8,	2.875189526
'''


import serial
import time
import serial.tools.list_ports

class GPD3303D:
    """
    Represents the GPD3303D Programmable Power Supply Unit (PSU).

    Args:
        emul (bool): Whether to enable emulation mode. Default is False.
        keyword (str): The keyword to search for in the device identification string. Default is 'GPD'.
        baudrate (int): The baud rate for serial communication. Default is 9600.
        timeout (float): The timeout duration for serial communication in seconds. Default is 1.
        port (str): The specific port to connect to. If not provided, the port will be searched automatically.

    Attributes:
        port (str): The port used for communication.
        keyword (str): The keyword used for device identification.
        timeout (float): The timeout duration for serial communication.
        baudrate (int): The baud rate for serial communication.
        inst (serial.Serial or str): The serial connection instance or a string indicating emulation mode.
        emul (bool): Whether emulation mode is enabled.
        inst_is_open (bool): Whether the serial connection is open.
        voltage (dict): A dictionary to store the voltage values for each channel.
        current (dict): A dictionary to store the current values for each channel.
        emul_str (str): A string indicating the emulation mode.

    Methods:
        connect(): Establishes a connection to the PSU.
        port_search(keyword): Searches for the PSU on available ports.
        send_cmd(cmd): Sends a command to the PSU and returns the response.
        get_idn(): Retrieves the identification string of the PSU.
        set_voltage(channel, voltage): Sets the voltage for a specific channel.
        set_current(channel, current): Sets the current for a specific channel.
        enable_output(): Enables the output of the PSU.
        disable_output(): Disables the output of the PSU.
        enable_beep(): Enables the beep sound of the PSU.
        disable_beep(): Disables the beep sound of the PSU.
        get_voltage(channel): Retrieves the voltage value for a specific channel.
        get_current(channel): Retrieves the current value for a specific channel.
        close_connection(): Closes the connection to the PSU.
        is_connected(): Checks if the connection to the PSU is established.
    """
    def __init__(self, emul=False, keyword='GPD', baudrate=9600, timeout=1, port=None) -> None:
        self.port = port
        self.keyword = keyword
        self.timeout = timeout
        self.baudrate = baudrate
        self.inst = None
        self.emul = emul
        self.inst_is_open = False
        self.voltage ={}
        self.current ={}
        if self.emul:
            self.emul_str = "[Emulation mode]:"
        else:
            self.emul_str = ""

    def connect(self):
        if not self.emul:
            if self.port:
                self.inst = serial.Serial(self.port, self.baudrate, timeout=self.timeout)
            else:
                self.inst = self.port_search(self.keyword)

            if self.inst is not None:
                print(f'Connected to {self.get_idn()}')
                return True
            else:
                print('PSU Connection failed.')
                return False
        else:
            print(f'{self.emul_str} Connected to GPD3303D PSU.')
            self.inst = 'emulated psu'
            self.inst_is_open = True
            return True


    def port_search(self, keyword):
        print('Searching for the device...')
        ports = serial.tools.list_ports.comports()
        for port, desc, hwid in sorted(ports):
            ser = serial.Serial(port, self.baudrate, timeout=self.timeout)
            ser.write(b'*IDN?\r\n')
            idn = ser.readline().strip().decode('ascii')
            
            if keyword in idn:
                print(f'"{keyword}" found in: {port}')
                return ser
            else:
                ser.close()
            
        print(f'"{keyword}" is not found on any port')
        return None

    
    def send_cmd(self, cmd):
        if self.is_connected():
            self.inst.write(cmd.encode('ascii') + b'\r\n')
            resp = self.inst.readline().strip().decode('ascii')
            return resp
        else:
            print('PSU Connection is not established.')
            return None

    
    def get_idn(self):
        if self.emul:
            return 'Emulated GPD3303D PSU'
        else:
            return self.send_cmd('*IDN?')
    
    def set_voltage(self, channel, voltage):
        if not self.emul:
            cmd = f'VSET{channel}:{voltage:.1f}'
            return self.send_cmd(cmd)
        else:
            self.voltage[channel] = voltage
            return True
    
    def set_current(self, channel, current):
        if not self.emul:
            cmd = f'ISET{channel}:{current:.2f}'
            return self.send_cmd(cmd)
        else:
            self.current[channel] = current
            return True
    
    def enable_output(self):
        if not self.emul: return self.send_cmd('OUT1')
        else:
            print(f'{self.emul_str} PSU output enabled.')
            return True

    
    def disable_output(self):
        if not self.emul: return self.send_cmd('OUT0')
        else:
            print(f'{self.emul_str} PSU output disabled.')
            return True
    
    def enable_beep(self):
        if not self.emul: return self.send_cmd('BEEP1')

    def disable_beep(self):
        if not self.emul: return self.send_cmd('BEEP0')
    
    def get_voltage(self, channel):
        if not self.emul:
            response = self.send_cmd(f'VOUT{channel}?')
            if response is not None:
                voltage_str, unit_str = response.strip().split("V")
                return float(voltage_str)
        else:
            return float(self.voltage[channel])

    
    def get_current(self, channel):
        if not self.emul:
            response = self.send_cmd(f'IOUT{channel}?')
            if response is not None:
                current_str, unit_str = response.strip().split("A")
                return float(current_str)
        else:
            return float(self.current[channel])
    
    def close_connection(self):
        if not self.emul:
            if self.inst is not None and self.inst.is_open:
                self.inst.close()
                print('PSU Connection closed.')
            else:
                print('No active PSU connection to close.')
        else:
            if self.inst is not None and self.inst_is_open:
                self.inst_is_open = False
                print(f'{self.emul_str} PSU Connection closed.')
            else:
                print(f'{self.emul_str} No active PSU connection to close.')
    
    def is_connected(self):
        if self.inst is None:
            return False
        if not self.emul: return self.inst.is_open
        else: return self.inst_is_open
#==============================================================================    

# how to use this class
if __name__ == '__main__':
    # create a power-supply object
    psu = GPD3303D(emul=True)
    #psu = GPD3303D(port='COM8')

    # connect to the device
    psu.connect()

    # Set the output voltage and current for channel 1
    psu.set_voltage(1, 12)
    psu.set_current(1, 0.01)

    # Set the output voltage and current for channel 2
    psu.set_voltage(2, 4.7)
    psu.set_current(2, 0.05)
    
    # Enable the output
    psu.enable_output()

    time.sleep(2)
    #print(f'voltage - CH1: {psu.get_voltage(1)} V')
    #print(f'current - CH1: {psu.get_current(1)} A')
    print(f'voltage - CH2: {psu.get_voltage(2)} V')
    #print(f'current - CH2: {psu.get_current(2)} A')
    
    # Disable the output
    psu.disable_output()

    # close the connection
    psu.close_connection()


import serial
import time
import serial.tools.list_ports
import logging

#----------------------------------------------
logging.getLogger().setLevel(logging.INFO)

class AFG2000:
    
    def __init__(self, emul=False, keyword='AFG', baudrate=9600, timeout=1, port=None) -> None:
        self.port = port
        self.keyword = keyword
        self.timeout = timeout
        self.baudrate = baudrate
        self.inst = None
    #---------------------------------------------------
    
    def connect(self):
        if self.inst is not None and self.inst.is_open:
            logging.info('AFG Connection is already established.')
            return True
        if self.port:
            self.inst = serial.Serial(self.port, self.baudrate, timeout=self.timeout)
        else:
            self.inst = self.port_search(self.keyword)

        if self.inst is not None:
            logging.info(f'Connected to {self.get_idn()}')
            return True
        else:
            logging.error('AFG Connection failed.')
            return False
    #---------------------------------------------------

    def port_search(self, keyword):
        logging.info('Searching for the device...')
        ports = serial.tools.list_ports.comports()
        for port, desc, hwid in sorted(ports):
            ser = serial.Serial(port, self.baudrate, timeout=self.timeout)
            ser.write(b'*IDN?\r\n')
            idn = ser.readline().strip().decode('ascii')
            
            if keyword in idn:
                logging.info(f'"{keyword}" found in: {port}')
                return ser
            else:
                ser.close()
            
        logging.info(f'"{keyword}" is not found on any port')
        return None
    #---------------------------------------------------

    def close(self):
        if self.inst is not None and self.inst.is_open:
            self.inst.close()
            logging.info('AFG Connection closed.')
        else:
            logging.info('No active AFG connection to close.')
    #---------------------------------------------------
    
    def is_connected(self):
        if self.inst is None:
            return False
        return self.inst.is_open
    #---------------------------------------------------
    
    def send_cmd(self, cmd):
        if self.is_connected():
            self.inst.write(cmd.encode('ascii') + b'\r\n')
            resp = self.inst.readline().strip().decode('ascii')
            return resp
        else:
            logging.info('AFG Connection is not established.')
            return None
    #---------------------------------------------------
            
    def get_idn(self):
        return self.send_cmd('*IDN?')
    #---------------------------------------------------
    
    def set_frequency(self, freq):
        """
        Sets the frequency of the AFG2000 function generator.

        Parameters:
        freq (float): The desired frequency in Hz.

        Returns:
        bool: True if the frequency is set successfully, False otherwise.
        """
        if self.is_connected():
            self.send_cmd(f'SOUR1:FREQ {freq}')
            return True
        else:
            logging.info('AFG2000 function generator is not connected!')
            return False
    	#---------------------------------------------------
	
    def set_waveform(self, waveform='SIN'):
        if self.is_connected():
            if waveform in ['SIN', 'SQU', 'RAMP', 'NOIS', 'ARB']:
                self.send_cmd(f'SOUR1:FUNC {waveform}')
                return True
            else:
                logging.error(f'Invalid waveform: {waveform}. Valid waveforms are "SIN", "SQU", "RAMP", "NOIS", and "ARB".')
                return None
        else:
            logging.info('AFG2000 function generator is not connected!')
            return False
	#---------------------------------------------------
        
    def set_amplitude(self, amplitude, unit='VPP'):
        if self.is_connected():
            if unit in ['VPP', 'VRMS', 'DBM']:
                self.send_cmd(f'SOUR1:VOLT:UNIT {unit}')
            else:
                logging.error(f'Invalid unit: {unit}. Valid units are "VPP", "VRMS", and "DBM".')
                return None
            self.send_cmd(f'SOUR1:AMPL {amplitude}')
            return True
        else:
            logging.info('AFG2000 function generator is not connected!')
            return False
    #---------------------------------------------------
        
    def set_DCoffset(self, offset):
        if self.is_connected():
            self.send_cmd(f'SOUR1:DCO {offset}')
            return True
        else:
            logging.info('AFG2000 function generator is not connected!')
            return False
    #---------------------------------------------------
        
    def set_output(self, state):
        if self.is_connected():
            if state in ['ON', 'OFF']:
                self.send_cmd(f'OUTP {state}')
                return True
            else:
                logging.error(f'Invalid output state: {state}. Valid states are "ON" and "OFF".')
                return None
        else:
            logging.info('AFG2000 function generator is not connected!')
            return False
#==============================================================================    

# how to use this class
if __name__ == '__main__':
    afg = AFG2000()
    afg.connect()
    afg.set_waveform('SIN')
    afg.set_frequency('70')
    afg.set_amplitude('0.2', 'VPP')
    afg.set_DCoffset('0.0')
    afg.set_output('ON')
    time.sleep(10)

    afg.set_amplitude('0.35')
    afg.set_frequency('90')
    time.sleep(10)
    afg.set_output('OFF')
    afg.close()


import os
import csv
import logging

#from YorkUphysLab.Actuator import Actuator as ACT
#from YorkUphysLab.HVcontrol import HV_control as HV
#----------------------------------------------

logging.getLogger().setLevel(logging.INFO)

def write_data_to_csv(data, path, filename):
    header = ['Position', 'Weight']
    
    # Create the full path for the CSV file on the desktop
    file_path = os.path.join(path, filename)
    
    try:
        with open(file_path, 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            csv_writer.writerow(header)
            
            for position, weight in data:
                csv_writer.writerow([position, weight])
        print(f"Data written to '{file_path}' successfully!")
        return True
    except Exception as e:
        print(f"An error occurred: {e}")
        return False
    
#----------------------------------------------
def reset_devices(actuator, hv):
    if not actuator.actuator_on:
        actuator.switch_on()
    actuator.set_position(0)

    if not hv.HV_on:
        hv.switch_on()
    hv.set_hv(0)

#----------------------------------------------
def multiply_lists(list_1, list_2):
    """
    Multiplies (mixes) two waveforms element-wise.
    
    Args:
        list_1 (list)
        list_2 (list)
    
    Returns:
        list: The resulting list after element-wise multiplication.
    """
    if len(list_1) != len(list_2):
        logging.error(f"Lists must be the same length: {len(list_1)} != {len(list_2)}")
        return None
    
    return [x*y for x,y in zip(list_1,list_2)]